#include <stdio.h>
#include <stdlib.h>
#include <netinet/ip.h>     /* struct iphdr */
#include <unistd.h>         /* read */
#include <sys/socket.h>
#include <string.h>         /* strerror */
#include <argp.h>
#include <stdbool.h>

#include "ospf.h"
#include "chksum.h"
#include "msgtype.h"
#include "authtype.h"
#include "print.h"
#include "init.h"
#include "getaddr.h"
#include "getiface.h"
#include "verbose.h"
#include "options.h"

const char * argp_program_bug_address = "wjholden@gmail.com";
const char * argp_program_version = "0.7";

#define PCKT_LEN 8192

bool quiet = false;
bool sendh = true;
bool verbose = false;
uint16_t helloInterval = 10;
uint32_t deadInterval = 40;
char * iface = NULL;
uint8_t priority = 0;
uint8_t options = 0;
uint8_t ttl = 1;
uint32_t area = 0;

/**************************** ARGP ****************************/
static int parse_opt (int key, char * arg, struct argp_state * state)
{
  char * err;
  
  switch (key)
    {
    case 'v': verbose = true; break;
    case 'q': quiet = true; break;
    case 's': sendh = false; break;
    case 'i':
      {
	helloInterval = strtoul(arg, &err, 0);
	if (* err != '\0')
	  argp_failure(state, 1, 0, "invalid hello interval");
	break;
      }
    case 'd':
      {
	deadInterval = strtoul(arg, &err, 0);
	if (* err != '\0')
	  argp_failure(state, 1, 0, "invalid dead interval");
	break;
      }
    case 'I': iface = arg; break;
    case 'p':
      {
	unsigned long int pri = strtoul(arg, &err, 0);
	if (pri > 255 || (* err) != '\0')
	  argp_failure(state, 1, 0, "invalid priority");
	priority = (uint8_t) pri;
	break;
      }
    case 't':
      {
	unsigned long int timetolive = strtoul(arg, &err, 0);
	if (timetolive > 255 || (* err) != '\0')
	  argp_failure(state, 1, 0, "invalid TTL value");
	ttl = (uint8_t) timetolive;
	break;
      }
    case 'D': options = options ^ OSPF_OPTION_DN; break;
    case 'O': options = options ^ OSPF_OPTION_O;  break;
    case 'C': options = options ^ OSPF_OPTION_DC; break;
    case 'L': options = options ^ OSPF_OPTION_L;  break;
    case 'P': options = options ^ OSPF_OPTION_NP; break;
    case 'M': options = options ^ OSPF_OPTION_MC; break;
    case 'E': options = options ^ OSPF_OPTION_E;  break;
    case 'T': options = options ^ OSPF_OPTION_MT; break;
    }
  return 0;
}

struct argp_option argument_options[] =
  {
    { 0, 0, 0, 0, "Program Options:", 5 },
    { "verbose", 'v', 0, 0, "explain what is being done", 5},
    { "quiet", 'q', 0, 0, "suppress packet header information", 5},
    { "simulate", 's', 0, 0, "don't actually send the hello packet", 5},

    { 0, 0, 0, 0, "Network Options:", 10 },
    { "interval", 'i', "seconds", 0, "hello interval (default: 10)", 10},
    { "dead", 'd', "seconds", 0, "router dead interval (default: 40)", 10},
    { "priority", 'p', "[0-255]", 0, "priority (default: 0)", 10},   
    { "interface", 'I', "interface", 0, "network interface", 10 },
    { "ttl", 't', "[0-255]", 0, "time to live (default: 1)", 10},

    { 0, 0, 0, 0, "OSPF Option Flags:", -10},
    { "dn", 'D', 0, 0, "set down bit for RFC 2547 BGP/MPLS VPNs", -10 },
    { "opaque", 'O', 0, 0, "set the O bit to support Opaque LSAs", -10 },
    { "dc", 'C', 0, 0, "set the DC bit to support Demand Circuits", -10 },
    { "lls", 'L', 0, 0, "set L bit for RFC 4813 Link-Local Signaling", -10 },
    { "nssa", 'N', 0, 0, "set N bit for Type-7 NSSA-External-LSAs", -10 },
    { "propogate", 'P', 0, 0, "set P bit to translate Type-7 LSA to Type-5", -10 },
    { "mospf", 'M', 0, 0, "set MC bit for Multicast extensions to OSPF", -10 },
    { "external", 'E', 0, 0, "clear E bit for external routing (set by default)", -10 },
    { "multitopology", 'T', 0, 0, "set MT bit for MT-OSPF", -10 },
    
    { 0 }
  };
/**************************** ARGP ****************************/

int main (int argc, char ** argv)
{
  struct iphdr * ip;
  struct OSPFheader * header;
  struct OSPFhello * hello;
  
  uint8_t buffer[PCKT_LEN];
  int sd;
  struct sockaddr_in sin;
  const int on = 1;

  options = OSPF_OPTION_E;

  uint32_t address, netmask;
  
  /**************************** ARGP ****************************/
  struct argp argp = { argument_options, parse_opt, "",
		       "Send an OSPF Hello Packet.\v"
		       "Use a packet capture utility, such as Wireshark, to view the response.", 0, 0, 0};
  argp_parse (&argp, argc, argv, 0, 0, 0);
  /**************************** ARGP ****************************/
  
  memset(buffer, 0, PCKT_LEN);

  ip = (struct iphdr *) buffer;
  header = (struct OSPFheader *) (buffer + sizeof(struct iphdr));
  hello = (struct OSPFhello *) (buffer + sizeof(struct iphdr) +
		 	        sizeof(struct OSPFheader));

  if (iface == NULL)
    iface = getDefaultInterface();

  address = getAddress(iface);
  netmask = getNetmask(iface);

  initIP(ip, address, ttl);
  initOSPFHeader(header, address, area);
  initOSPFHello(hello, address, netmask, helloInterval, deadInterval, priority, options);
  
  ip->tot_len = htons(sizeof(struct iphdr) + sizeof(struct OSPFheader) +
		      sizeof(struct OSPFhello));

  header->length = htons(sizeof(struct OSPFheader) + sizeof(struct OSPFhello));

  ip->check = 0;
  ip->check = chksum(ip, ip->ihl * 4); // interesting fact: do not change byte ordering

  header->chksum = 0;
  header->chksum = chksum(header, ntohs(header->length));

  if (!quiet)
    {
      printIP(ip);
      printOSPFHeader(header);
      printOSPFHello(hello);
    }
  
    if (sendh)
      {
  sin.sin_family = AF_INET;
  sin.sin_addr.s_addr = ip->daddr;

  sd = socket(PF_INET, SOCK_RAW, IPPROTO_OSPF);
  if (sd < 0)
    {
      perror("socket() failed");
      exit(EXIT_FAILURE);
    }

  if (setsockopt(sd, IPPROTO_IP, IP_HDRINCL, &on, sizeof(on)) < 0)
    {
      perror("setsockopt() failed to set IP_HDRINCL");
      exit(EXIT_FAILURE);
    }

  if (sendto(sd, buffer, ntohs(ip->tot_len), 0, (struct sockaddr *)&sin,
	     sizeof(struct sockaddr)) < 0)
    {
      perror("sendto() failed");
      exit(EXIT_FAILURE);
    }
  
  close(sd);
    }
  return(EXIT_SUCCESS);
}
